/*
 * kgraft_patch_bsc967773
 *
 * Fix for CVE-2016-2384, bsc#966693
 *
 *  Upstream commit:
 *  07d86ca93db7e5cdf4743564d98292042ec21af7
 *
 *  SLE12 commit:
 *  4b264846f792275a85156239eb709196cebcf15c
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/usb/audio.h>
#include <sound/core.h>
#include <sound/rawmidi.h>

#include "local.h"

#define KGR_PATCHED_MODULE	"snd-usbmidi-lib"

#if defined(CONFIG_SND_DEBUG) || defined(CONFIG_SND_VERBOSE_PRINTK)
__printf(4, 5)
static void (*kgr__snd_printk)(unsigned int level, const char *file, int line,
		  const char *format, ...);
#else
#define kgr__snd_printk(level, file, line, format, args...) \
	printk(format, ##args)
#endif

#define kgr_snd_printk(fmt, args...) \
	kgr__snd_printk(0, __FILE__, __LINE__, fmt, ##args)

#ifdef CONFIG_SND_DEBUG_VERBOSE
#define kgr_snd_printdd(format, args...) \
	kgr__snd_printk(2, __FILE__, __LINE__, format, ##args)
#else
__printf(1, 2)
static inline void kgr_snd_printdd(const char *format, ...) {}
#endif

#ifdef CONFIG_SND_DEBUG
#define kgr_snd_printd(fmt, args...) \
	kgr__snd_printk(1, __FILE__, __LINE__, fmt, ##args)
#else
__printf(1, 2)
static inline void kgr_snd_printd(const char *format, ...) {}
#endif

static struct usb_protocol_ops *kgr_snd_usbmidi_standard_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_maudio_broken_running_status_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_122l_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_midiman_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_novation_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_raw_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_emagic_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_cme_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_akai_ops;
static struct usb_protocol_ops *kgr_snd_usbmidi_ftdi_ops;
static struct snd_rawmidi_global_ops *kgr_snd_usbmidi_ops;
static struct snd_rawmidi_ops *kgr_snd_usbmidi_output_ops;
static struct snd_rawmidi_ops *kgr_snd_usbmidi_input_ops;
static void (*kgr_snd_usbmidi_error_timer)(unsigned long data);
static int (*kgr_snd_usbmidi_detect_endpoints)(struct snd_usb_midi* umidi, struct snd_usb_midi_endpoint_info* endpoint, int max_endpoints);
static int (*kgr_snd_usbmidi_detect_per_port_endpoints)(struct snd_usb_midi* umidi, struct snd_usb_midi_endpoint_info* endpoints);
static int (*kgr_snd_rawmidi_new)(struct snd_card *card, char *id, int device, int output_count, int input_count, struct snd_rawmidi ** rrawmidi);
static void (*kgr_snd_usbmidi_rawmidi_free)(struct snd_rawmidi *rmidi);
static void (*kgr_snd_usbmidi_out_endpoint_clear)(struct snd_usb_midi_out_endpoint *ep);
static void (*kgr_snd_usbmidi_out_urb_complete)(struct urb* urb);
static void (*kgr_snd_usbmidi_out_tasklet)(unsigned long data);
static void (*kgr_snd_usbmidi_in_endpoint_delete)(struct snd_usb_midi_in_endpoint* ep);
static void (*kgr_snd_usbmidi_in_urb_complete)(struct urb* urb);
static void (*kgr_snd_usbmidi_init_substream)(struct snd_usb_midi* umidi, int stream, int number, struct snd_rawmidi_substream ** rsubstream);
static void (*kgr_snd_rawmidi_set_ops)(struct snd_rawmidi *rmidi, int stream, struct snd_rawmidi_ops *ops);
static int (*kgr_usb_set_interface)(struct usb_device *dev, int interface, int alternate);
static int (*kgr_usb_control_msg)(struct usb_device *dev, unsigned int pipe, __u8 request, __u8 requesttype, __u16 value, __u16 index, void *data, __u16 size, int timeout);
static void (*kgr_usb_autopm_get_interface_no_resume)(struct usb_interface *intf);
static void *(*kgr_usb_alloc_coherent)(struct usb_device *dev, size_t size, gfp_t mem_flags, dma_addr_t *dma);
static struct urb *(*kgr_usb_alloc_urb)(int iso_packets, gfp_t mem_flags);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "snd_usbmidi_standard_ops", (void *)&kgr_snd_usbmidi_standard_ops },
	{ "snd_usbmidi_maudio_broken_running_status_ops", (void *)&kgr_snd_usbmidi_maudio_broken_running_status_ops },
	{ "snd_usbmidi_122l_ops", (void *)&kgr_snd_usbmidi_122l_ops },
	{ "snd_usbmidi_midiman_ops", (void *)&kgr_snd_usbmidi_midiman_ops },
	{ "snd_usbmidi_novation_ops", (void *)&kgr_snd_usbmidi_novation_ops },
	{ "snd_usbmidi_error_timer", (void *)&kgr_snd_usbmidi_error_timer },
	{ "snd_usbmidi_raw_ops", (void *)&kgr_snd_usbmidi_raw_ops },
	{ "snd_usbmidi_emagic_ops", (void *)&kgr_snd_usbmidi_emagic_ops },
	{ "snd_usbmidi_cme_ops", (void *)&kgr_snd_usbmidi_cme_ops },
	{ "snd_usbmidi_akai_ops", (void *)&kgr_snd_usbmidi_akai_ops },
	{ "snd_usbmidi_ftdi_ops", (void *)&kgr_snd_usbmidi_ftdi_ops },
	{ "snd_usbmidi_ops", (void *)&kgr_snd_usbmidi_ops },
	{ "snd_usbmidi_output_ops", (void *)&kgr_snd_usbmidi_output_ops },
	{ "snd_usbmidi_input_ops", (void *)&kgr_snd_usbmidi_input_ops },
	{ "snd_usbmidi_detect_endpoints", (void *)&kgr_snd_usbmidi_detect_endpoints },
	{ "snd_usbmidi_detect_per_port_endpoints", (void *)&kgr_snd_usbmidi_detect_per_port_endpoints },
	{ "snd_rawmidi_new", (void *)&kgr_snd_rawmidi_new },
	{ "snd_usbmidi_rawmidi_free", (void *)&kgr_snd_usbmidi_rawmidi_free },
	{ "snd_usbmidi_out_endpoint_clear", (void *)&kgr_snd_usbmidi_out_endpoint_clear },
	{ "snd_usbmidi_out_urb_complete", (void *)&kgr_snd_usbmidi_out_urb_complete },
	{ "snd_usbmidi_out_tasklet", (void *)&kgr_snd_usbmidi_out_tasklet },
	{ "snd_usbmidi_in_endpoint_delete", (void *)&kgr_snd_usbmidi_in_endpoint_delete },
	{ "snd_usbmidi_in_urb_complete", (void *)&kgr_snd_usbmidi_in_urb_complete },
	{ "snd_usbmidi_init_substream", (void *)&kgr_snd_usbmidi_init_substream },
	{ "snd_rawmidi_set_ops", (void *)&kgr_snd_rawmidi_set_ops },
	{ "usb_set_interface", (void *)&kgr_usb_set_interface },
	{ "usb_control_msg", (void *)&kgr_usb_control_msg },
	{ "usb_autopm_get_interface_no_resume", (void *)&kgr_usb_autopm_get_interface_no_resume },
	{ "usb_alloc_coherent", (void *)&kgr_usb_alloc_coherent },
	{ "usb_alloc_urb", (void *)&kgr_usb_alloc_urb },
#if defined(CONFIG_SND_DEBUG) || defined(CONFIG_SND_VERBOSE_PRINTK)
	{ "__snd_printk", (void *)&kgr__snd_printk },
#endif
};

/* isra optimized function */
static int kgr_snd_usbmidi_get_ms_info(struct snd_usb_midi* umidi,
	struct snd_usb_midi_endpoint_info* endpoints)
{
	struct usb_interface* intf;
	struct usb_host_interface *hostif;
	struct usb_interface_descriptor* intfd;
	struct usb_ms_header_descriptor* ms_header;
	struct usb_host_endpoint *hostep;
	struct usb_endpoint_descriptor* ep;
	struct usb_ms_endpoint_descriptor* ms_ep;
	int i, epidx;

	intf = umidi->iface;
	if (!intf)
		return -ENXIO;
	hostif = &intf->altsetting[0];
	intfd = get_iface_desc(hostif);
	ms_header = (struct usb_ms_header_descriptor*)hostif->extra;
	if (hostif->extralen >= 7 &&
	    ms_header->bLength >= 7 &&
	    ms_header->bDescriptorType == USB_DT_CS_INTERFACE &&
	    ms_header->bDescriptorSubtype == UAC_HEADER)
		kgr_snd_printdd(KERN_INFO "MIDIStreaming version %02x.%02x\n",
			    ms_header->bcdMSC[1], ms_header->bcdMSC[0]);
	else
		kgr_snd_printk(KERN_WARNING "MIDIStreaming interface descriptor not found\n");

	epidx = 0;
	for (i = 0; i < intfd->bNumEndpoints; ++i) {
		hostep = &hostif->endpoint[i];
		ep = get_ep_desc(hostep);
		if (!usb_endpoint_xfer_bulk(ep) && !usb_endpoint_xfer_int(ep))
			continue;
		ms_ep = (struct usb_ms_endpoint_descriptor*)hostep->extra;
		if (hostep->extralen < 4 ||
		    ms_ep->bLength < 4 ||
		    ms_ep->bDescriptorType != USB_DT_CS_ENDPOINT ||
		    ms_ep->bDescriptorSubtype != UAC_MS_GENERAL)
			continue;
		if (usb_endpoint_dir_out(ep)) {
			if (endpoints[epidx].out_ep) {
				if (++epidx >= MIDI_MAX_ENDPOINTS) {
					kgr_snd_printk(KERN_WARNING "too many endpoints\n");
					break;
				}
			}
			endpoints[epidx].out_ep = usb_endpoint_num(ep);
			if (usb_endpoint_xfer_int(ep))
				endpoints[epidx].out_interval = ep->bInterval;
			else if (snd_usb_get_speed(umidi->dev) == USB_SPEED_LOW)
				/*
				 * Low speed bulk transfers don't exist, so
				 * force interrupt transfers for devices like
				 * ESI MIDI Mate that try to use them anyway.
				 */
				endpoints[epidx].out_interval = 1;
			endpoints[epidx].out_cables = (1 << ms_ep->bNumEmbMIDIJack) - 1;
			kgr_snd_printdd(KERN_INFO "EP %02X: %d jack(s)\n",
				    ep->bEndpointAddress, ms_ep->bNumEmbMIDIJack);
		} else {
			if (endpoints[epidx].in_ep) {
				if (++epidx >= MIDI_MAX_ENDPOINTS) {
					kgr_snd_printk(KERN_WARNING "too many endpoints\n");
					break;
				}
			}
			endpoints[epidx].in_ep = usb_endpoint_num(ep);
			if (usb_endpoint_xfer_int(ep))
				endpoints[epidx].in_interval = ep->bInterval;
			else if (snd_usb_get_speed(umidi->dev) == USB_SPEED_LOW)
				endpoints[epidx].in_interval = 1;
			endpoints[epidx].in_cables = (1 << ms_ep->bNumEmbMIDIJack) - 1;
			kgr_snd_printdd(KERN_INFO "EP %02X: %d jack(s)\n",
				    ep->bEndpointAddress, ms_ep->bNumEmbMIDIJack);
		}
	}
	return 0;
}

/* Inlined */
static int kgr_snd_usbmidi_detect_yamaha(struct snd_usb_midi* umidi,
	struct snd_usb_midi_endpoint_info* endpoint)
{
	struct usb_interface* intf;
	struct usb_host_interface *hostif;
	struct usb_interface_descriptor* intfd;
	uint8_t* cs_desc;

	intf = umidi->iface;
	if (!intf)
		return -ENOENT;
	hostif = intf->altsetting;
	intfd = get_iface_desc(hostif);
	if (intfd->bNumEndpoints < 1)
		return -ENOENT;

	/*
	 * For each port there is one MIDI_IN/OUT_JACK descriptor, not
	 * necessarily with any useful contents.  So simply count 'em.
	 */
	for (cs_desc = hostif->extra;
	     cs_desc < hostif->extra + hostif->extralen && cs_desc[0] >= 2;
	     cs_desc += cs_desc[0]) {
		if (cs_desc[1] == USB_DT_CS_INTERFACE) {
			if (cs_desc[2] == UAC_MIDI_IN_JACK)
				endpoint->in_cables = (endpoint->in_cables << 1) | 1;
			else if (cs_desc[2] == UAC_MIDI_OUT_JACK)
				endpoint->out_cables = (endpoint->out_cables << 1) | 1;
		}
	}
	if (!endpoint->in_cables && !endpoint->out_cables)
		return -ENOENT;

	return kgr_snd_usbmidi_detect_endpoints(umidi, endpoint, 1);
}

/* Inlined */
static int kgr_snd_usbmidi_detect_roland(struct snd_usb_midi* umidi,
	struct snd_usb_midi_endpoint_info* endpoint)
{
	struct usb_interface* intf;
	struct usb_host_interface *hostif;
	u8* cs_desc;

	intf = umidi->iface;
	if (!intf)
		return -ENOENT;
	hostif = intf->altsetting;
	/*
	 * Some devices have a descriptor <06 24 F1 02 <inputs> <outputs>>,
	 * some have standard class descriptors, or both kinds, or neither.
	 */
	for (cs_desc = hostif->extra;
	     cs_desc < hostif->extra + hostif->extralen && cs_desc[0] >= 2;
	     cs_desc += cs_desc[0]) {
		if (cs_desc[0] >= 6 &&
		    cs_desc[1] == USB_DT_CS_INTERFACE &&
		    cs_desc[2] == 0xf1 &&
		    cs_desc[3] == 0x02) {
			endpoint->in_cables  = (1 << cs_desc[4]) - 1;
			endpoint->out_cables = (1 << cs_desc[5]) - 1;
			return kgr_snd_usbmidi_detect_endpoints(umidi, endpoint, 1);
		} else if (cs_desc[0] >= 7 &&
			   cs_desc[1] == USB_DT_CS_INTERFACE &&
			   cs_desc[2] == UAC_HEADER) {
			return kgr_snd_usbmidi_get_ms_info(umidi, endpoint);
		}
	}

	return -ENODEV;
}

/* Inlined */
static int kgr_snd_usbmidi_create_rawmidi(struct snd_usb_midi* umidi,
	int out_ports, int in_ports)
{
	struct snd_rawmidi *rmidi;
	int err;

	err = kgr_snd_rawmidi_new(umidi->card, "USB MIDI",
			      umidi->next_midi_device++,
			      out_ports, in_ports, &rmidi);
	if (err < 0)
		return err;
	strcpy(rmidi->name, umidi->card->shortname);
	rmidi->info_flags = SNDRV_RAWMIDI_INFO_OUTPUT |
			    SNDRV_RAWMIDI_INFO_INPUT |
			    SNDRV_RAWMIDI_INFO_DUPLEX;
	rmidi->ops = kgr_snd_usbmidi_ops;
	rmidi->private_data = umidi;
	rmidi->private_free = kgr_snd_usbmidi_rawmidi_free;
	kgr_snd_rawmidi_set_ops(rmidi, SNDRV_RAWMIDI_STREAM_OUTPUT, kgr_snd_usbmidi_output_ops);
	kgr_snd_rawmidi_set_ops(rmidi, SNDRV_RAWMIDI_STREAM_INPUT, kgr_snd_usbmidi_input_ops);

	umidi->rmidi = rmidi;
	return 0;
}

/* inlined */
static void kgr_snd_usbmidi_out_endpoint_delete(struct snd_usb_midi_out_endpoint *ep)
{
	kgr_snd_usbmidi_out_endpoint_clear(ep);
	kfree(ep);
}

/* isra optimized */
static int kgr_snd_usbmidi_out_endpoint_create(struct snd_usb_midi* umidi,
	struct snd_usb_midi_endpoint_info* ep_info,
	struct snd_usb_midi_endpoint* rep)
{
	struct snd_usb_midi_out_endpoint* ep;
	unsigned int i;
	unsigned int pipe;
	void* buffer;

	rep->out = NULL;
	ep = kzalloc(sizeof(*ep), GFP_KERNEL);
	if (!ep)
		return -ENOMEM;
	ep->umidi = umidi;

	for (i = 0; i < OUTPUT_URBS; ++i) {
		ep->urbs[i].urb = kgr_usb_alloc_urb(0, GFP_KERNEL);
		if (!ep->urbs[i].urb) {
			kgr_snd_usbmidi_out_endpoint_delete(ep);
			return -ENOMEM;
		}
		ep->urbs[i].ep = ep;
	}
	if (ep_info->out_interval)
		pipe = usb_sndintpipe(umidi->dev, ep_info->out_ep);
	else
		pipe = usb_sndbulkpipe(umidi->dev, ep_info->out_ep);
	switch (umidi->usb_id) {
	default:
		ep->max_transfer = usb_maxpacket(umidi->dev, pipe, 1);
		break;
		/*
		 * Various chips declare a packet size larger than 4 bytes, but
		 * do not actually work with larger packets:
		 */
	case USB_ID(0x0a92, 0x1020): /* ESI M4U */
	case USB_ID(0x1430, 0x474b): /* RedOctane GH MIDI INTERFACE */
	case USB_ID(0x15ca, 0x0101): /* Textech USB Midi Cable */
	case USB_ID(0x15ca, 0x1806): /* Textech USB Midi Cable */
	case USB_ID(0x1a86, 0x752d): /* QinHeng CH345 "USB2.0-MIDI" */
	case USB_ID(0xfc08, 0x0101): /* Unknown vendor Cable */
		ep->max_transfer = 4;
		break;
		/*
		 * Some devices only work with 9 bytes packet size:
		 */
	case USB_ID(0x0644, 0x800E): /* Tascam US-122L */
	case USB_ID(0x0644, 0x800F): /* Tascam US-144 */
		ep->max_transfer = 9;
		break;
	}
	for (i = 0; i < OUTPUT_URBS; ++i) {
		buffer = kgr_usb_alloc_coherent(umidi->dev,
					    ep->max_transfer, GFP_KERNEL,
					    &ep->urbs[i].urb->transfer_dma);
		if (!buffer) {
			kgr_snd_usbmidi_out_endpoint_delete(ep);
			return -ENOMEM;
		}
		if (ep_info->out_interval)
			usb_fill_int_urb(ep->urbs[i].urb, umidi->dev,
					 pipe, buffer, ep->max_transfer,
					 kgr_snd_usbmidi_out_urb_complete,
					 &ep->urbs[i], ep_info->out_interval);
		else
			usb_fill_bulk_urb(ep->urbs[i].urb, umidi->dev,
					  pipe, buffer, ep->max_transfer,
					  kgr_snd_usbmidi_out_urb_complete,
					  &ep->urbs[i]);
		ep->urbs[i].urb->transfer_flags = URB_NO_TRANSFER_DMA_MAP;
	}

	spin_lock_init(&ep->buffer_lock);
	tasklet_init(&ep->tasklet, kgr_snd_usbmidi_out_tasklet, (unsigned long)ep);
	init_waitqueue_head(&ep->drain_wait);

	for (i = 0; i < 0x10; ++i)
		if (ep_info->out_cables & (1 << i)) {
			ep->ports[i].ep = ep;
			ep->ports[i].cable = i << 4;
		}

	if (umidi->usb_protocol_ops->init_out_endpoint)
		umidi->usb_protocol_ops->init_out_endpoint(ep);

	rep->out = ep;
	return 0;
}

/* isra optimized */
static int kgr_snd_usbmidi_in_endpoint_create(struct snd_usb_midi* umidi,
	struct snd_usb_midi_endpoint_info* ep_info,
	struct snd_usb_midi_endpoint* rep)
{
	struct snd_usb_midi_in_endpoint* ep;
	void* buffer;
	unsigned int pipe;
	int length;
	unsigned int i;

	rep->in = NULL;
	ep = kzalloc(sizeof(*ep), GFP_KERNEL);
	if (!ep)
		return -ENOMEM;
	ep->umidi = umidi;

	for (i = 0; i < INPUT_URBS; ++i) {
		ep->urbs[i] = kgr_usb_alloc_urb(0, GFP_KERNEL);
		if (!ep->urbs[i]) {
			kgr_snd_usbmidi_in_endpoint_delete(ep);
			return -ENOMEM;
		}
	}
	if (ep_info->in_interval)
		pipe = usb_rcvintpipe(umidi->dev, ep_info->in_ep);
	else
		pipe = usb_rcvbulkpipe(umidi->dev, ep_info->in_ep);
	length = usb_maxpacket(umidi->dev, pipe, 0);
	for (i = 0; i < INPUT_URBS; ++i) {
		buffer = kgr_usb_alloc_coherent(umidi->dev, length, GFP_KERNEL,
					    &ep->urbs[i]->transfer_dma);
		if (!buffer) {
			kgr_snd_usbmidi_in_endpoint_delete(ep);
			return -ENOMEM;
		}
		if (ep_info->in_interval)
			usb_fill_int_urb(ep->urbs[i], umidi->dev,
					 pipe, buffer, length,
					 kgr_snd_usbmidi_in_urb_complete,
					 ep, ep_info->in_interval);
		else
			usb_fill_bulk_urb(ep->urbs[i], umidi->dev,
					  pipe, buffer, length,
					  kgr_snd_usbmidi_in_urb_complete, ep);
		ep->urbs[i]->transfer_flags = URB_NO_TRANSFER_DMA_MAP;
	}

	rep->in = ep;
	return 0;
}

/* Inlined */
static int kgr_snd_usbmidi_create_endpoints_midiman(struct snd_usb_midi* umidi,
	struct snd_usb_midi_endpoint_info* endpoint)
{
	struct snd_usb_midi_endpoint_info ep_info;
	struct usb_interface* intf;
	struct usb_host_interface *hostif;
	struct usb_interface_descriptor* intfd;
	struct usb_endpoint_descriptor* epd;
	int cable, err;

	intf = umidi->iface;
	if (!intf)
		return -ENOENT;
	hostif = intf->altsetting;
	intfd = get_iface_desc(hostif);
	/*
	 * The various MidiSport devices have more or less random endpoint
	 * numbers, so we have to identify the endpoints by their index in
	 * the descriptor array, like the driver for that other OS does.
	 *
	 * There is one interrupt input endpoint for all input ports, one
	 * bulk output endpoint for even-numbered ports, and one for odd-
	 * numbered ports.  Both bulk output endpoints have corresponding
	 * input bulk endpoints (at indices 1 and 3) which aren't used.
	 */
	if (intfd->bNumEndpoints < (endpoint->out_cables > 0x0001 ? 5 : 3)) {
		kgr_snd_printdd(KERN_ERR "not enough endpoints\n");
		return -ENOENT;
	}

	epd = get_endpoint(hostif, 0);
	if (!usb_endpoint_dir_in(epd) || !usb_endpoint_xfer_int(epd)) {
		kgr_snd_printdd(KERN_ERR "endpoint[0] isn't interrupt\n");
		return -ENXIO;
	}
	epd = get_endpoint(hostif, 2);
	if (!usb_endpoint_dir_out(epd) || !usb_endpoint_xfer_bulk(epd)) {
		kgr_snd_printdd(KERN_ERR "endpoint[2] isn't bulk output\n");
		return -ENXIO;
	}
	if (endpoint->out_cables > 0x0001) {
		epd = get_endpoint(hostif, 4);
		if (!usb_endpoint_dir_out(epd) ||
		    !usb_endpoint_xfer_bulk(epd)) {
			kgr_snd_printdd(KERN_ERR "endpoint[4] isn't bulk output\n");
			return -ENXIO;
		}
	}

	ep_info.out_ep = get_endpoint(hostif, 2)->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;
	ep_info.out_interval = 0;
	ep_info.out_cables = endpoint->out_cables & 0x5555;
	err = kgr_snd_usbmidi_out_endpoint_create(umidi, &ep_info, &umidi->endpoints[0]);
	if (err < 0)
		return err;

	ep_info.in_ep = get_endpoint(hostif, 0)->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;
	ep_info.in_interval = get_endpoint(hostif, 0)->bInterval;
	ep_info.in_cables = endpoint->in_cables;
	err = kgr_snd_usbmidi_in_endpoint_create(umidi, &ep_info, &umidi->endpoints[0]);
	if (err < 0)
		return err;

	if (endpoint->out_cables > 0x0001) {
		ep_info.out_ep = get_endpoint(hostif, 4)->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;
		ep_info.out_cables = endpoint->out_cables & 0xaaaa;
		err = kgr_snd_usbmidi_out_endpoint_create(umidi, &ep_info, &umidi->endpoints[1]);
		if (err < 0)
			return err;
	}

	for (cable = 0; cable < 0x10; ++cable) {
		if (endpoint->out_cables & (1 << cable))
			kgr_snd_usbmidi_init_substream(umidi, SNDRV_RAWMIDI_STREAM_OUTPUT, cable,
						   &umidi->endpoints[cable & 1].out->ports[cable].substream);
		if (endpoint->in_cables & (1 << cable))
			kgr_snd_usbmidi_init_substream(umidi, SNDRV_RAWMIDI_STREAM_INPUT, cable,
						   &umidi->endpoints[0].in->ports[cable].substream);
	}
	return 0;
}

/* Inlined */
static int kgr_snd_usbmidi_create_endpoints(struct snd_usb_midi* umidi,
	struct snd_usb_midi_endpoint_info* endpoints)
{
	int i, j, err;
	int out_ports = 0, in_ports = 0;

	for (i = 0; i < MIDI_MAX_ENDPOINTS; ++i) {
		if (endpoints[i].out_cables) {
			err = kgr_snd_usbmidi_out_endpoint_create(umidi, &endpoints[i],
							      &umidi->endpoints[i]);
			if (err < 0)
				return err;
		}
		if (endpoints[i].in_cables) {
			err = kgr_snd_usbmidi_in_endpoint_create(umidi, &endpoints[i],
							     &umidi->endpoints[i]);
			if (err < 0)
				return err;
		}

		for (j = 0; j < 0x10; ++j) {
			if (endpoints[i].out_cables & (1 << j)) {
				kgr_snd_usbmidi_init_substream(umidi, SNDRV_RAWMIDI_STREAM_OUTPUT, out_ports,
							   &umidi->endpoints[i].out->ports[j].substream);
				++out_ports;
			}
			if (endpoints[i].in_cables & (1 << j)) {
				kgr_snd_usbmidi_init_substream(umidi, SNDRV_RAWMIDI_STREAM_INPUT, in_ports,
							   &umidi->endpoints[i].in->ports[j].substream);
				++in_ports;
			}
		}
	}
	kgr_snd_printdd(KERN_INFO "created %d output and %d input ports\n",
		    out_ports, in_ports);
	return 0;
}

/* Patched function */
int kgr_snd_usbmidi_create(struct snd_card *card, struct usb_interface* iface,
	struct list_head *midi_list, const struct snd_usb_audio_quirk* quirk)
{
	struct snd_usb_midi* umidi;
	struct snd_usb_midi_endpoint_info endpoints[MIDI_MAX_ENDPOINTS];
	int out_ports, in_ports;
	int i, err;

	umidi = kzalloc(sizeof(*umidi), GFP_KERNEL);
	if (!umidi)
		return -ENOMEM;
	umidi->dev = interface_to_usbdev(iface);
	umidi->card = card;
	umidi->iface = iface;
	umidi->quirk = quirk;
	umidi->usb_protocol_ops = kgr_snd_usbmidi_standard_ops;
	init_timer(&umidi->error_timer);
	spin_lock_init(&umidi->disc_lock);
	init_rwsem(&umidi->disc_rwsem);
	mutex_init(&umidi->mutex);
	umidi->usb_id = USB_ID(le16_to_cpu(umidi->dev->descriptor.idVendor),
			       le16_to_cpu(umidi->dev->descriptor.idProduct));
	umidi->error_timer.function = kgr_snd_usbmidi_error_timer;
	umidi->error_timer.data = (unsigned long)umidi;

	/* detect the endpoint(s) to use */
	memset(endpoints, 0, sizeof(endpoints));
	switch (quirk ? quirk->type : QUIRK_MIDI_STANDARD_INTERFACE) {
	case QUIRK_MIDI_STANDARD_INTERFACE:
		err = kgr_snd_usbmidi_get_ms_info(umidi, endpoints);
		if (umidi->usb_id == USB_ID(0x0763, 0x0150)) /* M-Audio Uno */
			umidi->usb_protocol_ops =
				kgr_snd_usbmidi_maudio_broken_running_status_ops;
		break;
	case QUIRK_MIDI_US122L:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_122l_ops;
		/* fall through */
	case QUIRK_MIDI_FIXED_ENDPOINT:
		memcpy(&endpoints[0], quirk->data,
		       sizeof(struct snd_usb_midi_endpoint_info));
		err = kgr_snd_usbmidi_detect_endpoints(umidi, &endpoints[0], 1);
		break;
	case QUIRK_MIDI_YAMAHA:
		err = kgr_snd_usbmidi_detect_yamaha(umidi, &endpoints[0]);
		break;
	case QUIRK_MIDI_ROLAND:
		err = kgr_snd_usbmidi_detect_roland(umidi, &endpoints[0]);
		break;
	case QUIRK_MIDI_MIDIMAN:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_midiman_ops;
		memcpy(&endpoints[0], quirk->data,
		       sizeof(struct snd_usb_midi_endpoint_info));
		err = 0;
		break;
	case QUIRK_MIDI_NOVATION:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_novation_ops;
		err = kgr_snd_usbmidi_detect_per_port_endpoints(umidi, endpoints);
		break;
	case QUIRK_MIDI_RAW_BYTES:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_raw_ops;
		/*
		 * Interface 1 contains isochronous endpoints, but with the same
		 * numbers as in interface 0.  Since it is interface 1 that the
		 * USB core has most recently seen, these descriptors are now
		 * associated with the endpoint numbers.  This will foul up our
		 * attempts to submit bulk/interrupt URBs to the endpoints in
		 * interface 0, so we have to make sure that the USB core looks
		 * again at interface 0 by calling usb_set_interface() on it.
		 */
		if (umidi->usb_id == USB_ID(0x07fd, 0x0001)) /* MOTU Fastlane */
			kgr_usb_set_interface(umidi->dev, 0, 0);
		err = kgr_snd_usbmidi_detect_per_port_endpoints(umidi, endpoints);
		break;
	case QUIRK_MIDI_EMAGIC:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_emagic_ops;
		memcpy(&endpoints[0], quirk->data,
		       sizeof(struct snd_usb_midi_endpoint_info));
		err = kgr_snd_usbmidi_detect_endpoints(umidi, &endpoints[0], 1);
		break;
	case QUIRK_MIDI_CME:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_cme_ops;
		err = kgr_snd_usbmidi_detect_per_port_endpoints(umidi, endpoints);
		break;
	case QUIRK_MIDI_AKAI:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_akai_ops;
		err = kgr_snd_usbmidi_detect_per_port_endpoints(umidi, endpoints);
		/* endpoint 1 is input-only */
		endpoints[1].out_cables = 0;
		break;
	case QUIRK_MIDI_FTDI:
		umidi->usb_protocol_ops = kgr_snd_usbmidi_ftdi_ops;

		/* set baud rate to 31250 (48 MHz / 16 / 96) */
		err = kgr_usb_control_msg(umidi->dev, usb_sndctrlpipe(umidi->dev, 0),
				      3, 0x40, 0x60, 0, NULL, 0, 1000);
		if (err < 0)
			break;

		err = kgr_snd_usbmidi_detect_per_port_endpoints(umidi, endpoints);
		break;
	default:
		kgr_snd_printd(KERN_ERR "invalid quirk type %d\n", quirk->type);
		err = -ENXIO;
		break;
	}
	if (err < 0) {
		kfree(umidi);
		return err;
	}

	/* create rawmidi device */
	out_ports = 0;
	in_ports = 0;
	for (i = 0; i < MIDI_MAX_ENDPOINTS; ++i) {
		out_ports += hweight16(endpoints[i].out_cables);
		in_ports += hweight16(endpoints[i].in_cables);
	}
	err = kgr_snd_usbmidi_create_rawmidi(umidi, out_ports, in_ports);
	if (err < 0) {
		kfree(umidi);
		return err;
	}

	/* create endpoint/port structures */
	if (quirk && quirk->type == QUIRK_MIDI_MIDIMAN)
		err = kgr_snd_usbmidi_create_endpoints_midiman(umidi, &endpoints[0]);
	else
		err = kgr_snd_usbmidi_create_endpoints(umidi, endpoints);
	if (err < 0) {
		/*
		 * Fix CVE-2016-2384
		 *  -snd_usbmidi_free(umidi);
		 */
		return err;
	}

	kgr_usb_autopm_get_interface_no_resume(umidi->iface);

	list_add_tail(&umidi->list, midi_list);
	return 0;
}

static int kgr_patch_bsc967773_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc967773_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc967773_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc967773_module_nb = {
	.notifier_call = kgr_patch_bsc967773_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc967773_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc967773_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc967773_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc967773_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc967773_module_nb);
}
